//
// ControlWindow Implemention
//
// this belongs to the 'Tower of Benoi' Project
// Copyright 1998-1999
//
// Written by: Marco Zinn
//

#ifndef	CONTROL_WINDOW_H
#define	CONTROL_WINDOW_H
//#include "TowerWindow.h"

#include <InterfaceKit.h> // for the BSlider
//#include <string.h>

// Forward declaration for DiscTowerView class. You'll find the definition in TowerWindow.h
class DiscTowerView; 

// Definition of ControlWindow class
//  This window controls the auto-solving-process
//  As it has no other use, it is only visible, when auto-solving is running.
//  It contains a slider which determins the speed of the solving (delay between moves)
//  and a Button to stop the whole process.
class ControlWindow : public BWindow {
	public:
		ControlWindow(BRect frame, const char* name,DiscTowerView *FromTower, DiscTowerView *ToTower, DiscTowerView *BufferTower);
		void Reset();
		void StartSolving();
	private:
		status_t SolveGame();
			// Starts the solving process
		void StopSolving();
			// Stops it
		status_t BuildTower(DiscTowerView *ToTower, int NumDiscs);
			// Recursively called routine used for solving
		status_t MoveDiscTo(DiscTowerView *ToTower, int DiscNr);
			// Routine to move a single disc to a specific tower
		BSlider* SpeedSlider;
			// Pointer to the Slider
		BButton* ToggleButton;
		bool StopFlag;
			// Flag for stopping the solving process
		static int32 thread_func(void *arg); // STATIC function which is spawned as a new thread.
		thread_id	SolvingThread; // id of the solving thread (when spawned)
		DiscTowerView *From;
		DiscTowerView *To;
		DiscTowerView *Buffer;
};
 
#endif /* CONTROL_WINDOW_H */ 