//
// Implementation of Disc, DiscTowerView and TowerWindow classes
//
// this belongs to the 'Tower of Benoi' Project
// Copyright 1998-1999
//
// Written by: Marco Zinn
//

#include <InterfaceKit.h>
#include <stdio.h>
#include "TowerWindow.h"
#include "HistoryWindow.h"
#include "Towers.h"

// Implementation of Disc class

// The constructor saves the size and color and initializes the string
// according to the size
Disc::Disc(int newSize, rgb_color newColor)
		: BStringItem(""), size(newSize), color(newColor) {
	char name[10];
	sprintf (name,"%i", size);
	//	printf("New Disc: %i\n",size); // Debug
	BStringItem::SetText(name);
}

// DrawItem
//  Draws the disc, that means the colored bar and the label in white ('1'...'10')
//  DummyDiscs are drawn as 1-Pixel Lines making up a 'pole',

void Disc::DrawItem(BView *owner, BRect frame, bool complete) { 
	// if (this==DummyDisc) return; // Don't draw the Dummy-Discs. (Comment out for DEBUGing)

	// printf("Disc %i draws\n",size); // Debug
	
	// If complete drawing is required, erase the whole background
	//  Note that we don't care, if the item is selected. This would be the place to do that
	if (complete) {
		owner->SetHighColor(owner->ViewColor());
    	owner->FillRect(frame); 
    	}   	
 
	// Do some calculations about the position of the bar:
	//  Take the frame rectangle (which is the area for the whole Item to draw in),
	//  calc the center (average x) and center the disc around that
	BRect original;
	original=frame;
	frame.left=((original.left+original.right)/2)-size*10;
	frame.right=((original.left+original.right)/2)+size*10;

	owner->SetHighColor(color);	// Select the Disc's color
    owner->FillRect(frame);		// Draw the Disc
	
	if (this==DummyDisc) return; // Don't label the Dummy-Discs.

	owner->SetLowColor(color);	// Treat the color as LowColor (AKA Background color!)
	owner->SetHighColor(255,255,255); // Set Drawing color to white
	owner->MovePenTo((frame.left+frame.right)/2-3, frame.bottom-2); // Position pen somewhere near the center of the disc 
	owner->DrawString(Text()); // Draw text (label)
	// Note: While drawing a text, be sure the Low color is set to the background color
	//       This is important, because the text is anti-aliased using the High- and Low-
	//       Color, not the actual background on screen! (again: AFAIK!)
}

// Implementation of DiscTowerView class

// The constructor essantially saves the necessary data and Resets the Tower
DiscTowerView::DiscTowerView(BRect frame, int InitTowerID)
			: BListView(frame, "DiscTowerView",B_SINGLE_SELECTION_LIST, B_FOLLOW_ALL_SIDES, B_WILL_DRAW ) {
	// printf("New DiscTowerView\n"); // Debug
	myTowerID=InitTowerID;
	ResetDiscs(); // Initialize Tower
}


// InitiateDrag is called, when a drag-and-drop action is started, ie when the
// user drags something 'off' the Tower. This builds the nice little outline
// rectangle and contructs the OFFER_DISC message.
bool DiscTowerView::InitiateDrag(BPoint a, int32 index, bool wasSelected) {
	// printf("InitiateDrag: FirstDisc=%i\n",FirstDisc); // Debug
	if (TowerEmpty()) return false; // ERROR-Check
	
	be_app->PostMessage(STOP_SOLVING);

	if (acquire_sem(SolvingSem) != B_NO_ERROR) return false; // Wait for the solver thread to exit
	release_sem(SolvingSem);
		
	printf("Doing Drag\n"); // Debug
	
	// Construct outline Box:
	int DS=FirstDiscPtr()->size; 
		// Fetch the size of the uppermost disc, as only the uppermost disc will move at all!

	BPoint b=a; // a is the Point where the user klicked
	a.x-=DS*20; a.y-=FirstDiscPtr()->Height(); // This creates a rect, where the mouse pointer is at the lower right corner

	BMessage messy(OFFER_DISC); // Contruct a message
	messy.AddPointer("DiscPtr",FirstDiscPtr()); // Add a Ptr to the uppermost Disc. That's the disc we want to offer (move)
	
//	BRect rect(0.0, 0.0, DS*20, FirstDiscPtr()->Height()); 
//	BBitmap *image = new BBitmap(rect, B_RGB32,true);
//	printf("Bitmap constructed\n"); // Debug
//
//	BView *bitmapView = new BView(rect,"BitMapView",B_FOLLOW_NONE,B_WILL_DRAW);
//	printf("BitMapView constructed\n"); // Debug
//	
////	bitmapView->SetHighColor(FirstDiscPtr()->color);	// Select the Disc's color
//	bitmapView->SetHighColor(255, 0, 0, 96); //
////	bitmapView->SetLowColor(255, 0, 0, 96); //transparent white (for the highlight box)
//    bitmapView->FillRect(rect);		// Draw the Disc
//	bitmapView->SetDrawingMode(B_OP_ALPHA);
//	bitmapView->SetBlendingMode(B_PIXEL_ALPHA, B_ALPHA_OVERLAY);
//	bitmapView->SetViewColor(B_TRANSPARENT_32_BIT);
//	
////	bitmapView->SetLowColor(FirstDiscPtr()->color);	// Treat the color as LowColor (AKA Background color!)
////	bitmapView->SetHighColor(255,255,255); // Set Drawing color to white
////	bitmapView->MovePenTo((rect.left+rect.right)/2-3, rect.bottom-2); // Position pen somewhere near the center of the disc 
////	bitmapView->DrawString(FirstDiscPtr()->Text()); // Draw text (label)
//	printf("BitMapView drawn\n"); // Debug
//	bitmapView->Flush();
//	printf("BitMapView sync'ed\n"); // Debug
//	image->AddChild(bitmapView);
//	printf("BitMapView added\n"); // Debug
//	
//	delete bitmapView;

//	DragMessage(&messy,image, BPoint(DS*20, FirstDiscPtr()->Height())); // Start Drag operation.
	DragMessage(&messy,BRect(a,b)); // Start Drag operation.
	return true; // Say: 'It's ok. keep dragging away....'
	// Note: I never had to do such a thing in Windooze.
	// But I guess, you have to love the concept. Dragging in three lines!
	// The app_server does the rest for us (except dealing the incoming messages. see below)
	// Also note, that the app_server (or some other wizard in the depth of the system)
	// makes a copy of the message we pass to DragMessage(). The message we created
	// is just a 'model message', which can be deleted after the call to Dragmessage.
	// [I didn't wanna explain the whole messaging system...]
}


// ResetDiscs clear the Towers at center and to the right
// and stacks the left one full of Discs.
// The Locking is important. (AFAIK, but it costs a lot of crashed, if you don't lock!)
void DiscTowerView::ResetDiscs() {
	if (Window()) Window()->Lock(); // If the View is attached to a Window, Lock that
	MakeEmpty(); // Clear the List

	if (myTowerID==LEFT_TOWER){ // for left Tower
		for (int DiscCount=1;DiscCount<=NumberOfDiscs;DiscCount++) // loop starts at 1
				AddItem(Discs[DiscCount],DiscCount-1); // Stack the Discs, smallest first (and uppermost)
		FirstDisc=0; 
			// And mark the Tower as full. The First Disc is the uppermost (index 0)
		}
	else { // For the other Towers
		for (int DiscCount=1;DiscCount<=NumberOfDiscs;DiscCount++)
				AddItem(DummyDisc,DiscCount-1); // Stack DummyDiscs !
		FirstDisc=NumberOfDiscs; 
			// Declare the Tower as empty. Note that you shouldn't look, what is at
			// index 'FirstDisc', as this will be NULL.
		}

	if (Window()) Window()->Unlock(); // Unlock the Window, if the View is attached.
	// Note to Locking: 
	//  AFAIK (but you may look at the BeBook, if you like), you have to lock the window,
	//  as soon as you change the view hierarchie. By adding and removing Items (this 
	//  absolutely includes MakeEmpty()! ), you do exactly this, as the seperate Items in
	//  the list are Views that are children of the BListView.
	//  [Anybody correct me, if I'm wrong!]
}		

// MessageReceived does, what it alway does: it reacts to the messages it knows
// and passed the other 'up' to BListView
void DiscTowerView::MessageReceived(BMessage *message){
	switch(message->what) {
		case OFFER_DISC: // Somebody offers a disc to us.
			{
			// printf("Disc has been offered.\n"); // Debug
			
			// First check, if it's from another application. This is just a 
			// play-around-feature. ;-)
			if (message->IsSourceRemote()) {
			   	BAlert *alert = new BAlert("Warning!","No Cheating Please!","All right"); 
				alert->Go(); // Oh, by the way: Never delete the BAlert object!
				break;
				}

			// For further examination, find out, what the other has to offer.
			Disc *DiscPtr;
			message->FindPointer("DiscPtr",(void **)(&DiscPtr)); // get the Pointer out of the incoming message

			if (HasItem(DiscPtr)) { // We already have that disc. So return
				// Note: This happens, when you drop on the same tower where you dragged from
				// ... or, it's a bug ;-)
				//	printf("Disc is already here.\n"); // Debug
				break;
				}

			bool accept = false; // Decision flag. Do we accept the disc?

			if (TowerEmpty()) // An empty Tower gladly accepts everthing
				{
				// printf("Tower ist empty. Accepted.\n"); // Debug
				accept = true;
				}
			else // Tower is not empty. Compare the sizes of the discs.
				if ((FirstDiscPtr()->size) > DiscPtr->size)
					{
					// Accept, if the offered (dragged) disc is smaller than the one on top
					// printf("Other Disc is bigger. Accepted.\n"); // Debug
					accept = true;
					}

			if (accept) {
				(FirstDisc)--; // We'll soon have one more disc, so move upward
				//	printf("FirstDisc here is now %i\n",FirstDisc); // Debug
				if (FirstDisc<0) { printf("ERROR: FirstDisc<0!!!!\n"); break; }; // Error-Check!

				// Make up a DISC_ACCEPTED reply for the offering Tower
				//  Include 'myName', because the offering Tower will have to
				//  make the report to the history window.
				BMessage myReply(DISC_ACCEPTED);
				myReply.AddInt32("RecTower",myTowerID);
				message->SendReply(&myReply); // Send it back to where it came from
						
				// After the message has gone, put the offered disc on top.
				ReplaceItem(FirstDisc,DiscPtr);
				// Note: FirstDisc points to the 'empty' place on top of the tower
				// Although this place seems empty, it contained a DummyDisc.
				// Through the ReplaceItem call, that DummyDisc is removed and the offered
				// disc is placed at it's position.
						
				// Check, if the user has won
				if ((myTowerID != LEFT_TOWER) && (TowerFull())) {
				   	BAlert *alert = new BAlert("Info","You did it !","Really?"); 
					alert->Go(NULL); 
						// Go(NULL) opens an asnychronous Alert Box. 
						// We can't wait here, cause that would cause an ugly effect:
						// The last disc exists twice right now, because the offering
						// tower has not yet deleted it. And we don't want the user
						// to see that, do we?
						// Anyway, we don't care, when he presses the button of the Alert
						// Box. He will do it.... some time.
					be_app->PostMessage(STOP_SOLVING); // just in case auto-solving was running: resets labels etc
					}
				}
			else { // that means: accepted=false
				// Deny the offer. Probably the move is not allowed.
				// printf("Other Disc is smaller. DENIED.\n"); // Debug
				BMessage myreply(DISC_DENIED); // Make up a simple reply.
				message->SendReply(&myreply); // send it back.
				}
			}
		break;

		case DISC_ACCEPTED: 
		// We get a reply message: we offered a disc to some other tower, and that
		// tower accepted:
			{
			int32 RecTower; 
			// printf("Original Tower: Offered Disc has been accepted.\n"); // Debug
			message->FindInt32("RecTower",(&RecTower)); // gets the Pointer to the String!
			// Note: I have no clue, where the string is stored. Probably, it's stored
			//  in the message object, which is not valid for a long time.
			//  The (holy) BeBook tells me to copy the string.... It works without that,
			//  'cause I only need a character for about one millisecond ;-)
			
			int DS=FirstDiscPtr()->size; 
			// Save the size of the uppermost Disc. Needed for the History display
			
			ReplaceItem(FirstDisc,DummyDisc);
			// Replace the uppermost Disc with a DummyDisc
			//  This 'removes' the Disc on top. For more, see above.

			FirstDisc++; // The tower is a bit smaller now, so the uppermost disc is one position 'deeper'
			// printf("Original Tower: Disc has been removed.\n"); // Debug
			// printf("Original Tower: FirstDisc here is now %i\n",FirstDisc); // Debug

			// As the move is done now, show it:
			((TowerApp*)be_app)->History->AddMove(myTowerID,RecTower,DS);
			// This sends the two labels of the participating towers and the size of
			// the Disc to the HistoryWindow. That will inc the counter and display
			// the move.
			}
		break;
		case DISC_DENIED:
		// We get an other reply message: we offered a disc to some other tower,
		// but that tower didn't like it. Inform the user about the desaster
			{
		   	BAlert *alert = new BAlert("Info!","You can't play a Disc\n on a smaller one!","Too bad..."); 
			alert->Go();
			}
		break;
	default: // all other messages are passed to BListView
		BListView::MessageReceived(message);
		break;
	}
}


// Implementation of TowerWindow class

// The Contructor sets of the Window (uh....)
//  That one constructs the menu (with the submenu, thus it's a bit lengthy),
//  and adds the three towers and their labels as children.
TowerWindow::TowerWindow(BRect frame, const char* name, HistoryWindow* History)
			: BWindow(frame, name, B_TITLED_WINDOW,B_NOT_ZOOMABLE|B_NOT_H_RESIZABLE|B_WILL_ACCEPT_FIRST_CLICK) {

	// printf("New TowerWindow\n"); // Debug
	
	// The whole stuff below constructs a menu with submenu
	BMenuBar* mb = new BMenuBar(BRect(0,0,frame.right,15), "menubar"); // the menu bar
	BMenu* gameMenu = new BMenu("Game");  // the game menu
	BMenu* rMenu = new BMenu("Reset"); // the Reset submenu
	
	// Contruction of the submenu including construction of the model messages:
	// Contruct model message for resetting the game with the same number of discs (0 means 'don't change the number discs, just reset')
	BMessage* ResetWithSameNumberMessy = new BMessage(RESET_GAME);
	ResetWithSameNumberMessy->AddInt32("Discs",0);
	rMenu->AddItem(new BMenuItem("Keep Number of Discs", ResetWithSameNumberMessy,'R'));
	rMenu->AddSeparatorItem();

	// Contruct model message for resetting the game with 2 discs
	BMessage* ResetMessy2 = new BMessage(RESET_GAME);
	ResetMessy2->AddInt32("Discs",2);
	rMenu->AddItem(new BMenuItem("Use 2 Discs", ResetMessy2,'2'));
	// Contruct model message for resetting the game with 3 discs
	BMessage* ResetMessy3 = new BMessage(RESET_GAME);
	ResetMessy3->AddInt32("Discs",3);
	rMenu->AddItem(new BMenuItem("Use 3 Discs", ResetMessy3,'3'));
	// Contruct model message for resetting the game with 4 discs
	BMessage* ResetMessy4 = new BMessage(RESET_GAME);
	ResetMessy4->AddInt32("Discs",4);
	rMenu->AddItem(new BMenuItem("Use 4 Discs", ResetMessy4,'4'));
	// Contruct model message for resetting the game with 5 discs
	BMessage* ResetMessy5 = new BMessage(RESET_GAME);
	ResetMessy5->AddInt32("Discs",5);
	rMenu->AddItem(new BMenuItem("Use 5 Discs", ResetMessy5,'5'));
	// Contruct model message for resetting the game with 6 discs
	BMessage* ResetMessy6 = new BMessage(RESET_GAME);
	ResetMessy6->AddInt32("Discs",6);
	rMenu->AddItem(new BMenuItem("Use 6 Discs", ResetMessy6,'6'));
	// Contruct model message for resetting the game with 7 discs
	BMessage* ResetMessy7 = new BMessage(RESET_GAME);
	ResetMessy7->AddInt32("Discs",7);
	rMenu->AddItem(new BMenuItem("Use 7 Discs", ResetMessy7,'7'));
	// Contruct model message for resetting the game with 8 discs
	BMessage* ResetMessy8 = new BMessage(RESET_GAME);
	ResetMessy8->AddInt32("Discs",8);
	rMenu->AddItem(new BMenuItem("Use 8 Discs", ResetMessy8,'8'));
	// Contruct model message for resetting the game with 9 discs
	BMessage* ResetMessy9 = new BMessage(RESET_GAME);
	ResetMessy9->AddInt32("Discs",9);
	rMenu->AddItem(new BMenuItem("Use 9 Discs", ResetMessy9,'9'));
	// Contruct model message for resetting the game with 10 discs
	BMessage* ResetMessy10 = new BMessage(RESET_GAME);
	ResetMessy10->AddInt32("Discs",10);
	rMenu->AddItem(new BMenuItem("Use 10 Discs", ResetMessy10,'0'));
	// (uh.oh. done with the submenu)

	gameMenu->AddItem(rMenu); // add the whole submenu to the menu
	SolveToggler = new BMenuItem("Start Auto-Solve", new BMessage(START_SOLVING),'S');
	gameMenu->AddItem(SolveToggler);
	gameMenu->AddSeparatorItem();
	gameMenu->AddItem(new BMenuItem("Quit", new BMessage(B_QUIT_REQUESTED),'Q'));

	// all messages should directly go to out app instead of the window
	gameMenu->SetTargetForItems(be_app);
	rMenu->SetTargetForItems(be_app);
	mb->AddItem(gameMenu); // add menu to bar

	// make the help  menu

	BMenu *helpMenu = new BMenu("?");  // the help menu
	helpMenu->AddItem(new BMenuItem("Documentation", new BMessage(SHOW_DOCS),'D'));
	helpMenu->AddItem(new BMenuItem("About", new BMessage(B_ABOUT_REQUESTED),'A'));
	helpMenu->SetTargetForItems(be_app);
	mb->AddItem(helpMenu); // add menu to bar

	AddChild(mb); // add the menu bar to the window
	
	// Add the Towers and their labels above them
	BRect r(0,0,200,150); // this will be used by the towers
	BRect r1(0,0,200,15); // and that by the labels

	r.OffsetBy(5,35); // adjust the position
	r1.OffsetBy(99,18); // that one, too
	LeftTowerView=new DiscTowerView(r,LEFT_TOWER); // Create Start Tower
	AddChild(LeftTowerView); // add it
	AddChild(new BStringView(r1,"label","A")); // add label
	// Note: we won't need the labels anymore, so we can forget about them at once
	// The towers (LeftTowerView, CenterTowerView, RightTowerView) are important to us. So, they are constructed
	// another way and the pointers are saved.
	// Also note, that (AFAIK) you don't have to bother about deleting the whole stuff,
	// this will be done, when the window is deleted, which is done, when the app
	// is deleted (or somewhen around that time)

	// allright, the whole stuff again for Towers B and C
	r.OffsetBy(200,0);
	r1.OffsetBy(200,0);
	CenterTowerView=new DiscTowerView(r,CENTER_TOWER);
	AddChild(CenterTowerView);
	AddChild(new BStringView(r1,"label","B"));

	r.OffsetBy(200,0);
	r1.OffsetBy(200,0);
	RightTowerView=new DiscTowerView(r,RIGHT_TOWER);
	AddChild(RightTowerView);
	AddChild(new BStringView(r1,"label","C"));

	Show(); // Show the Window
	// Note that we still are in the constructor of the application!
	// This might not be the best time to show() a window, but it works and the
	// others (other tutorials) do it the same way.
	// It might be better to show the windows when the app is READY_TO_RUN.
	// Try that is your app crashed at start ;-)
}

void TowerWindow::Reset() {
	// Reset the Towers.
	//  Note, that the Window-Locking-Part is done in ResetDiscs(), that means
	//  that the Window is locked and unlocked three times altogether. Not perfect, but it works
	LeftTowerView->ResetDiscs();
	CenterTowerView->ResetDiscs();
	RightTowerView->ResetDiscs();

	SolveToggler->SetLabel("Start Auto-Solve");
	SolveToggler->SetMessage(new BMessage(START_SOLVING));
	SolveToggler->SetShortcut('S',B_COMMAND_KEY);
}

void TowerWindow::MessageReceived(BMessage *message){
	switch(message->what) {
		case START_SOLVING: // got a request to solve the game
			SolveToggler->SetLabel("Stop Auto-Solve");
			SolveToggler->SetMessage(new BMessage(STOP_SOLVING));
			SolveToggler->SetShortcut('O',B_COMMAND_KEY);
			break;
		case STOP_SOLVING: // got a request to solve the game
			SolveToggler->SetLabel("Start Auto-Solve");
			SolveToggler->SetMessage(new BMessage(START_SOLVING));
			SolveToggler->SetShortcut('S',B_COMMAND_KEY);
			break;
		default: // Any other messages are passed to BApplication. DO NOT FORGET THIS!
			BWindow::MessageReceived(message);
			break;
	}
}



// QuitRequested is called, when the user wants to close the window.
//  Notifiy the application of that (so it may close down the rest) and allow the closing
bool TowerWindow::QuitRequested() {
	// printf("*** TowerWin::QuitReq\n"); // Debug
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}

 